
#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "estad.h"
#include "histos.h"
#include "imag.h"
#include "inclina0.h"
#include "opcns.h"

#define K64 65536
#define K32 32768

/*************    Inclinacion    *************/

void pon_vprim_vult2 (int inc, int tu, int tv, int *vprim, int *vult)
{
  *vprim = 0;
  *vult  = tv - 1;
  if (inc > 0) *vult  -= (inc * tu + K32) / K64;
  if (inc < 0) *vprim -= (inc * tu - K32) / K64;
}

int ponSignok32 (int inc)
{
  if (inc >= 0)
    return  K32;
  else
    return -K32;
}

void histVert (int *hisV, int *nxp, int *nxu, tvent vent, int inc)
{
  int tx, ty;
  char ** pix;
  int x, y, xp, xu, xi, ac;
  int redondeo;
  tx  = vent->tx ;
  ty  = vent->ty ;
  pix = vent->pix ;
  redondeo = ponSignok32 (inc);
  pon_vprim_vult (inc, ty, tx, &xp, &xu) ;
  for (x=xp; x <= xu; x++){
    ac = 0;
    for (y=0; y < ty; y++){
      xi = x + (y * inc + redondeo) / K64 ;
      ac += (pix [y][xi] & 0xFF) ;
      }
    hisV[x] = ac ;
    }
  *nxp=xp; *nxu=xu;
}

double fondoHistVert (tvent vent, int inc)
{
  int hisV[MAXPIXH];
  int xp, xu;
  histVert (hisV, &xp, &xu, vent, inc);
  return 255.0 - minv (&hisV[xp], xu-xp+1) / vent->ty ;
}

static double desviacionHistV (tvent vent, int inc)
{
  int xp, xu;
  int hisV[MAXPIXH];
  histVert (hisV, &xp, &xu, vent, inc);
  return  desviacion (&hisV[xp], xu-xp+1) / vent->ty ;
}

static double desviacionHistVCuad (tvent vent, int inc)
{
  int tx, ty;
  int xp, xu;
  int hisV[MAXPIXH];
  tx  = vent->tx ;
  ty  = vent->ty ;
  pon_vprim_vult (inc, ty, tx, &xp, &xu) ;
  histoVer3InclCuad (hisV, vent, inc);
  return  desviacion (&hisV[xp], xu-xp+1) / ty ;
}

double promedioHistVCuad (tvent vent, int inc)
{
  int tx, ty;
  int xp, xu;
  int hisV[MAXPIXH];
  tx  = vent->tx ;
  ty  = vent->ty ;
  pon_vprim_vult (inc, ty, tx, &xp, &xu) ;
  histoVer3InclCuad (hisV, vent, inc);
  return  maximoD (&hisV[xp], xu-xp+1) ;
}

double desviacionHistV2 (tvent vent, int inc)
{
  int hisV[MAXPIXH];
  int xp, xu;
  histoVer2Incl (hisV, vent, inc) ;
  pon_vprim_vult (inc, vent->ty, vent->tx, &xp, &xu) ;
  return  desviacion (&hisV[xp], xu-xp+1) ;
}

/* mientras no se usa ...
static double desviacionHistH (tvent vent, int inc)
{
  int tx, ty;
  char ** pix;
  int x, y, yp, yu, yi, imas1, ac;
  int hisH[MAXPIXV];
  tx  = vent->tx ;
  ty  = vent->ty ;
  pix = vent->pix ;
  pon_cp_cu_imas1 (inc, ty, &yp, &yu, &imas1) ;
  for (y=yp; y <= yu; y++){
    ac = 0;
    for (x=0; x < tx; x++){
      yi = y + (x * imas1) / tx ;
      ac += (pix [yi][x] & 0xFF) ;
      }
    hisH[y] = ac ;
    }
  return  desviacion (&hisH[yp], yu-yp+1) / tx ;
}
*/

#define INCSAL  5

static double desviacionHistHrap (tvent vent, int inc)
{
  int tx, ty;
  char ** pix;
  int x, y, yp, yu, yi, ac;
  int redondeo;
  int hisH[MAXPIXV];
  tx  = vent->tx ;
  ty  = vent->ty ;
  pix = vent->pix ;
  redondeo = ponSignok32 (inc);
  pon_vprim_vult (inc, tx, ty, &yp, &yu);
  for (y=yp; y <= yu; y++){
    ac = 0;
    for (x=0; x < tx; x += INCSAL){
      yi = y + (x * inc + redondeo) / K64 ;
      ac += (pix [yi][x] & 0xFF) ;
      }
    hisH[y] = ac ;
    }
  return  desviacion (&hisH[yp], yu-yp+1) / (tx / INCSAL) ;
}

void desviaciones (tvent vent)
{
  int inc;
  for(inc = -40*256; inc <= 40*256; inc+=256)
    printf ("desviacion: %3d %12.3f\n", inc, desviacionHistHrap(vent, inc));
}

/* para que no sean mayores que el tamanyo correspondiente */
void corrigeInclinacionesLim (int *incmnc, int *incmxc, tvent vent, char * nomFun, int incmn, int incmx)
  {
  int medintincl;   // medio del intervalo
  // int incmnc, incmxc;  /* inclinaciones corregidas, no superan t(xy) */
  int tamu, tamv;
  if (nomFun[0] == 'H') {
    tamu = vent->tx;
    tamv = vent->ty;
    }
  else if (nomFun[0] == 'V') {
    tamu = vent->ty;
    tamv = vent->tx;
    }
  else {
    printf ("ocre: inclina.calculaInclinacion nombre de funcion incorrecto: %s\n",
             nomFun);
    exit (1);
    }
  medintincl = (incmn + incmx) / 2;
  if ((2*tamv/3) < (medintincl*tamu/K64)) {
    printf ("ocre: inclina.calculaInclinacion intervalo incorrecto %d %d %d %d\n",
             tamu, tamv, incmn, incmx);
    exit (1);
    }
  while ((2*tamv/3) < (max(incmx,-incmn)*tamu/K64)) {
      incmx = (medintincl + incmx) / 2;
      incmn = (medintincl + incmn) / 2;
      }
  *incmxc = incmx;
  *incmnc = incmn;
  }

// incmx - incmn  es pontencia de 2 
int calculaInclinacion (tvent vent, double desviacionHist(tvent, int), char * nomFun, int incmn, int incmx)
{
  int inc, incop0, incop;
  double dt, dtx;
  int deltaInc ;
  int del0, del1;
  int incmnc, incmxc;   // inclinaciones corregidas, no superan t(xy) 
  corrigeInclinacionesLim (&incmnc, &incmxc, vent, nomFun, incmn, incmx);
// si del0 == 0 .. inclinacion >> medio 
  del0= (incmxc - incmnc) / 4;
  del1 = del0 / 2;
  incop = 0;
  dtx = 0.0 ;
  if (d_valori == 10)
    printf ("desviacion %s ", nomFun);
  for(inc = incmnc; inc <= incmxc; inc+=del0) {
    dt = desviacionHist(vent, inc) ;
    if (d_valori >= 20)
      printf ("desviacion %s %3d %12.3f\n", nomFun, inc, dt);
    if (d_valori == 10)
      printf (":");
    if (dt > dtx) {
      incop = inc;
      dtx = dt;
      }
    }
  for (deltaInc=del1; deltaInc>=1; deltaInc /=2){
    incop0 = incop;
    for (inc = incop0-deltaInc; inc <= incop0+deltaInc; inc+=(deltaInc*2)) {
      dt = desviacionHist(vent, inc) ;
      if (d_valori >= 20)
        printf ("desviacion %s %3d %12.3f\n", nomFun, inc, dt);
      if (d_valori == 10)
        printf (".");
      if (dt > dtx) {
        incop = inc;
        dtx = dt;
        }
      }
    }
  if (d_valori == 10)
    printf ("\n");
  if (d_valori >= 10)
    printf ("inclinacionOpt = %2d (calc)\n", incop) ;
  return incop;
}

int calculaInclinacionV (tvent vent)
{
  return calculaInclinacion (vent, desviacionHistV, "V", -2048, +2048);
}

int calculaInclinacionVfondo (tvent vent)
{
  return calculaInclinacion (vent, fondoHistVert, "V", -2048, +2048);
}

int calculaInclinacionH (tvent vent)
{
  // return calculaInclinacion (vent, desviacionHistHrap, "Hrap", -2048, +2048);
  return calculaInclinacion (vent, desviacionHistHrap, "Hrap", -4096, +4096);
}

int calculaInclinacionVCuad (tvent vent)
{
  return calculaInclinacion (vent, desviacionHistVCuad, "VCuad", -16384, +16384);
}

/*************    Giro    *************/

int despIncl (int inc, int x, int tx)
{
  if(inc > 0)
    return (x*inc+K32)/K64;
  else
    return (x*inc-K32)/K64;
}

void giroPeqH (tvent vent, int inc)
{
  int aux[MAXPIXV];
  int x, y, s;
  int tx, ty;
  char ** pix;
  tx =vent->tx ;
  ty =vent->ty ;
  pix = vent->pix ;
  for(x=0; x < tx; x++){
    if (inc>0){
      s= despIncl(inc,x,tx);
      for(y=0; y<s; y++)
        aux[y] = pix[y][x] ;
      for(y=s; y<ty; y++)
        pix[y-s][x] = pix[y][x] ;
      for (y=0; y<s; y++)
        pix[y+ty-s][x] = aux[y] ;
      }
    if (inc<0){
      s= - despIncl(inc,x,tx);
      for(y=ty-1;y>=(ty-s);y--)
        aux[y+s-ty] = pix[y][x] ;
      for(y=ty-s-1;y>=0;y--)
        pix[y+s][x] = pix[y][x] ;
      for(y=s-1;y>=0; y--)
        pix[y][x] = aux[y] ;
      }
    }
  if (d_valori >= 10)
    printf ("giro pequenyo H de %2d (pixels)\n", inc) ;
}

void giroPeqV (tvent vent, int inc)
{
  int aux[MAXPIXV];
  int x, y, s;
  int tx, ty;
  char ** pix;
  tx =vent->tx ;
  ty =vent->ty ;
  pix = vent->pix ;
  for(y=0; y < ty; y++){
    if (inc>0){
      s= despIncl(inc,y,ty);
      for(x=0; x<s; x++)
        aux[x] = pix[y][x] ;
      for(x=s; x<tx; x++)
        pix[y][x-s] = pix[y][x] ;
      for (x=0; x<s; x++)
        pix[y][x+tx-s] = aux[x] ;
      }
    if (inc<0){
      s= - despIncl(inc,y,ty);
      for(x=tx-1;x>=(tx-s);x--)
        aux[x+s-tx] = pix[y][x] ;
      for(x=tx-s-1;x>=0;x--)
        pix[y][x+s] = pix[y][x] ;
      for(x=s-1;x>=0; x--)
        pix[y][x] = aux[x] ;
      }
    }
  if (d_valori >= 10)
    printf ("giro pequenyo V de %2d (pixels)\n", inc) ;
}

