
#include <stdio.h>
#include <stdlib.h>

#include "base.h"
#include "lista.h"
#include "segmen.h"

// -------------------- segmento --------------------

tsegm nuevoSegm (int u0, int u1)
{
  tsegm segm;
  segm = (tsegm) malloc (sizeof(struct tesegm));
  segm -> u0 = u0 ;
  segm -> u1 = u1 ;
  return segm ;
}

tsegm copiaSegm (tsegm segm)
{
  return nuevoSegm (segm->u0, segm->u1);
}

void imprimeSegmento (tsegm segm)
{
  printf ("%4d %4d %4d\n",
	   segm->u0, segm->u1,
           segm->u1-segm->u0 + 1);
}

int tamanyoSegm (tsegm segm)
{
  return segm->u1 - segm->u0 + 1 ;
}

int tamanyoIntersegmSig (tnodol nodol)
{
  tsegm segm1, segm2;
  if (nodol->siguiente == 0L) {
     printf ("segmen.tamanyoIntersegmSig error:");
     imprimeSegmento ((tsegm) nodol->aitem);
     exit (10);
     }
  segm1 = (tsegm) nodol->aitem;
  segm2 = (tsegm) nodol->siguiente->aitem;
  return segm2->u0 - segm1->u1 - 1 ;
}



int centroSegm (tsegm segm)
{
  return ( segm->u1 + segm->u0 ) / 2 ;
}

// -------------------- lista de segmentos --------------------

void anyadeSegm (lista segmentos, int u0, int u1)
{
  tsegm segm;
  segm = nuevoSegm (u0, u1);
  aniadeafinlista (segmentos, (tsegm) segm);
}

void eliminaSegm (lista segmentos, tnodol nodol)
{
  eliminaNdelista (segmentos, nodol);
  free (nodol->aitem); 
  free (nodol);
}

void eliminaUltimoSegm (lista segmentos)
{
  eliminaSegm (segmentos, segmentos->ultimo);
}

void imprimeSegmentos (lista segmentos)
{
  int i=0;
  tnodol nodol;
  for (nodol = segmentos->primero; nodol != 0L;nodol = nodol->siguiente) {
     printf ("%4d ", i);
     imprimeSegmento ( (tsegm) nodol->aitem );
     i++ ;
     }
}

void imprimeHuecos (lista segmentos)
{
  int i=0;
  tnodol nodol;
  tsegm segm, segms;
  if (segmentos->primero != 0L || 
      segmentos->primero->siguiente != 0L) {
     for (nodol = segmentos->primero;
          nodol->siguiente != 0L;nodol = nodol->siguiente) {
        segm  = (tsegm) nodol->aitem;
        segms = (tsegm) nodol->siguiente->aitem;
        printf ("%4d %4d %4d %4d\n", i,
                segm->u1+1, segms->u0-1, segms->u0 - segm->u1 -1);
        i++ ;
      }
    }
}

void imprimeSegmentosyHuecos (lista segmentos)
{
  int i=0;
  tnodol nodol;
  tsegm segm, segms;
  if (segmentos->primero != 0L) {
     nodol = segmentos->primero;       // si solo hay un segmento
     if (segmentos->primero->siguiente != 0L) {
        for (nodol = segmentos->primero;
             nodol->siguiente != 0L;nodol = nodol->siguiente) {
          segm  = (tsegm) nodol->aitem;
          segms = (tsegm) nodol->siguiente->aitem;
          printf ("%4d %4d %4d %4d   ",
                   i, segm->u0, segm->u1, segm->u1 - segm->u0 +1);
          printf (" %4d %4d %4d\n",
                   segm->u1+1, segms->u0-1, segms->u0 - segm->u1 -1);
          i++ ;
          }
       }
       segm  = (tsegm) nodol->aitem;
       printf ("%4d %4d %4d %4d\n",
                i, segm->u0, segm->u1, segm->u1 - segm->u0 +1);
    }
}

void uneasegmentoSiguiente (lista segmentos, tnodol nodoseg)
{
  tsegm segm, segmSig;
  if (nodoseg->siguiente == 0L){
     printf ("segmen.uneasegmentoSiguiente: error , no hay siguiente \n");
     exit (99);
     }
  segm    = (tsegm) nodoseg->aitem;
  segmSig = (tsegm) nodoseg->siguiente->aitem;
  segm->u1 = segmSig->u1;
  eliminaSegm (segmentos, nodoseg->siguiente);
}

lista huecos (lista segmentos)
{
  tnodol nodol;
  tsegm segm, segms;
  lista lseg;
  lseg = listaNueva ();
  if (!listaVacia(segmentos) &&
      segmentos->primero->siguiente != 0L)
     for (nodol = segmentos->primero;
          nodol->siguiente != 0L;nodol = nodol->siguiente) {
        segm  = (tsegm) nodol->aitem;
        segms = (tsegm) nodol->siguiente->aitem;
        anyadeSegm (lseg, segm->u1+1, segms->u0-1);
       }
  return lseg;
}

void desplazaSegmentos (lista segmentos, int desp)
{
  tnodol nodol;
  tsegm segm;
  for (nodol = segmentos->primero;
       nodol != 0L;nodol = nodol->siguiente) {
     segm = (tsegm) nodol->aitem;
     segm->u0 += desp ;
     segm->u1 += desp ;
     }
}

void segmentosBajoUmbral (lista * nsegm, int *v, int tv, int umbral)
{
  lista segm;
  int i;
  int comienzo;
  segm = listaNueva();
  if (v[0] <  umbral) 
     comienzo = 0;
  for (i=0; i<tv-1; i++) {
     if ((v[i] >= umbral) && (v[i+1] <  umbral))
        comienzo = i+1;
     if ((v[i] <  umbral) && (v[i+1] >=  umbral))
        anyadeSegm (segm, comienzo, i+1);
     }
  if (v[tv-1] <  umbral) 
     anyadeSegm (segm, comienzo, tv-1);
  *nsegm = segm;
}

void segmentosSobreoenUmbral (lista * nsegm, int *v, int tv, int umbral)
{
  lista segm;
  int i;
  int comienzo;
  segm = listaNueva();
  if (v[0] >=  umbral) 
     comienzo = 0;
  for (i=0; i<tv-1; i++) {
     if ((v[i] <  umbral) && (v[i+1] >=  umbral))
        comienzo = i+1;
     if ((v[i] >= umbral) && (v[i+1] <   umbral))
        anyadeSegm (segm, comienzo, i+1);
     }
  if (v[tv-1] >= umbral) 
     anyadeSegm (segm, comienzo, tv-1);
  *nsegm = segm;
}

Boolean enSegmento (tsegm *vsegm, lista segmentos, int u)
{
  tnodol nodol;
  tsegm segm;
  Boolean encontrado;
  encontrado = FALSE;
  for (nodol = segmentos->primero;
       nodol != (tnodol)0L; nodol = nodol->siguiente) {
     segm = (tsegm) nodol->aitem;
     if ((segm->u0 <= u) && (u <= segm->u1)) {
        encontrado = TRUE;
        break;
        }
     }
  *vsegm = segm;
  return encontrado;
}

Boolean tocan11Segmentos (tsegm segm1, tsegm segm2)
{
  return  max(segm1->u0, segm2->u0) <=
          min(segm1->u1, segm2->u1) ;
}

tsegm union11Segmentos (tsegm segm1, tsegm segm2)
{
  return nuevoSegm ( min(segm1->u0, segm2->u0),
                     max(segm1->u1, segm2->u1) );
}

tsegm interseccion11Segmentos (tsegm segm1, tsegm segm2)
{
  return nuevoSegm ( max(segm1->u0, segm2->u0),
                     min(segm1->u1, segm2->u1) );
}

lista interseccionNnSegmentos (lista lsegms1, lista lsegms2)
{
  tnodol ns1, ns2;
  tsegm segm1, segm2;
  lista lsegms;
  lsegms = listaNueva ();
  for ( ns1 = lsegms1->primero, ns2 = lsegms2->primero;
       (ns1 != 0L && ns2 != 0L) ; ) {
     segm1 = (tsegm) ns1->aitem;
     segm2 = (tsegm) ns2->aitem;
     if (tocan11Segmentos(segm1, segm2))
        aniadeafinlista ( lsegms, interseccion11Segmentos(segm1, segm2) );
     if (segm1->u1 <= segm2->u1)
        ns1 = ns1->siguiente;
     if (segm2->u1 <= segm1->u1)
        ns2 = ns2->siguiente;
     }
  return lsegms;
}

// pasa dos listas a una lista ordenada por  u0
lista mezclaSegmentos (lista lsegms1, lista lsegms2)
{
  tnodol ns1, ns2;
  tsegm segm1, segm2;
  lista lsegms;
  lsegms = listaNueva ();
  // mezcla
  for ( ns1 = lsegms1->primero, ns2 = lsegms2->primero;
       (ns1 != 0L && ns2 != 0L) ; ) {
     segm1 = (tsegm) ns1->aitem;
     segm2 = (tsegm) ns2->aitem;
     if (segm1->u0 <= segm2->u0){
        ns1 = ns1->siguiente;
        aniadeafinlista (lsegms, segm1);
        }
     else {
        ns2 = ns2->siguiente;
        aniadeafinlista (lsegms, segm2);
        }
     }
  // resto de una lista
  if (ns1 != 0L)
     for ( ; ns1 != 0L; ns1 = ns1->siguiente) {
        segm1 = (tsegm) ns1->aitem;
        aniadeafinlista (lsegms, segm1);
       }
  else  //  (ns2 != 0L)
     for ( ; ns2 != 0L; ns2 = ns2->siguiente) {
        segm2 = (tsegm) ns2->aitem;
        aniadeafinlista (lsegms, segm2);
       }
  return lsegms;
}

// incluso con listas de segmentos que se solapen
lista eliminaHuecosMenores (lista lsegms, int umbr)
{
  tnodol ns;
  tsegm segmsig, segmn;
  lista lsegmslimp;
  lsegmslimp = listaNueva ();
  if (! listaVacia (lsegms)) {
    segmn = copiaSegm ((tsegm) lsegms->primero->aitem);
    for ( ns = lsegms->primero->siguiente; ns != 0L; ns=ns->siguiente ) {
       segmsig = (tsegm) ns->aitem;
       if ((segmn->u1 + 1 + umbr) >= segmsig->u0)
          segmn = union11Segmentos (segmn, segmsig);
       else {
          aniadeafinlista (lsegmslimp, segmn);
          segmn = copiaSegm ((tsegm) segmsig);
          }
       }
    aniadeafinlista (lsegmslimp, segmn);
    }
  return lsegmslimp;
}

lista reuneSegmentosMenores (lista lsegms, int umbr)  // con anterior si es posible
{
  tnodol ns;
  tsegm segmsig, segmn;
  lista lsegmslimp;
  lsegmslimp = listaNueva ();
  if (! listaVacia (lsegms)) {
    segmn = copiaSegm ((tsegm) lsegms->primero->aitem);
    for ( ns = lsegms->primero->siguiente; ns != 0L; ns=ns->siguiente ) {
       segmsig = (tsegm) ns->aitem;
       // if ((segmn->u1 + 1 + umbr) >= segmsig->u0)
       if ( ((segmn->u1   - segmn->u0   + 1) < umbr) ||
            ((segmsig->u1 - segmsig->u0 + 1) < umbr)   )
          segmn = union11Segmentos (segmn, segmsig);
       else {
          aniadeafinlista (lsegmslimp, segmn);
          segmn = copiaSegm ((tsegm) segmsig);
          }
       }
    aniadeafinlista (lsegmslimp, segmn);
    }
  return lsegmslimp;
}

lista unionNnSegmentos (lista lsegms1, lista lsegms2)
{
  lista lsegmsmez, lsegms;
  lsegmsmez = mezclaSegmentos (lsegms1, lsegms2);
  lsegms = eliminaHuecosMenores (lsegmsmez, 0);
  return lsegms;
}

// -------------------- especif. segmenta_palabras --------------------

void huecosUnionSegmentos (lista * ncortes, lista blosr, lista blosi, int alturaLin)
{
 lista blosr2, blosi2, blos;
 blosr2   = eliminaHuecosMenores (blosr, alturaLin/7);   // antes 5
 blosi2   = eliminaHuecosMenores (blosi, alturaLin/8);   // antes 6
 blos     = unionNnSegmentos (blosr2, blosi2);
 *ncortes = huecos (blos);
}

void huecosInterseccionSegmentos (lista * ncortes, lista blosr, lista blosi, int alturaLin)
{
 lista blosr2, blosi2, blos;
 blosr2   = eliminaHuecosMenores (blosr, alturaLin/9);   // antes 5 ,7
 blosi2   = eliminaHuecosMenores (blosi, alturaLin/8);   // antes 6
 blos     = interseccionNnSegmentos (blosr2, blosi2);
 *ncortes = huecos (blos);
}

/**/   // dudosa utilidad (+2 meses?)   mar may  3 20:52:01 CEST 2011
void intersecSegmentos2 (lista * ncortes, lista blosr, lista blosi, int alturaLin)
{
  lista cortes;
  tnodol nsa, nsb;
  Boolean fna, fnb; 
  int ca, cb, fa, fb, ci, fi;
  int umbralR;
  int umbralI;
  enum {avanzaA, avanzaB, avanzanAyB} avan;
  //      antes      5 y 6      luego      7 y 8
  umbralR =  alturaLin/7;
  umbralI =  alturaLin/8;
  cortes = listaNueva ();
  if ((blosr->primero != 0L) & (blosi->primero != 0L)) {
    nsa = blosr->primero;
    nsb = blosi->primero;
    fna = TRUE ;
    fnb = TRUE ;
    while (( nsa->siguiente != 0L) && (nsb->siguiente != 0L)) {
      ca = ((tsegm)nsa->aitem)->u1 + 1 ;  fa = ((tsegm)nsa->siguiente->aitem)->u0 - 1 ;
      cb = ((tsegm)nsb->aitem)->u1 + 1 ;  fb = ((tsegm)nsb->siguiente->aitem)->u0 - 1 ;
      if      (cb > fa)  avan = avanzaA ;
      else if (ca > fb)  avan = avanzaB ;
      else {
        ci = max (ca, cb) ;           fi = min (fa, fb) ;
        if  ( (fa-ca > umbralR-1) ||
              (fb-cb > umbralI-1) ) {
	  if (fna && fnb) {
            anyadeSegm (cortes, ci, fi) ;
	    fna = FALSE ;  fnb = FALSE ;
	    }
	  else if ( (fi-ci+1) > tamanyoSegm((tsegm)cortes->ultimo->aitem) ) {
	    eliminaUltimoSegm (cortes) ;
            anyadeSegm (cortes, ci, fi) ;
	    fna = FALSE ;  fnb = FALSE ;
	    }
          }
	  // else nada
        if      (fa < fb) avan = avanzaA ;
        else if (fb < fa) avan = avanzaB ;
        else              avan = avanzanAyB ; // fa == fb 
        }
        switch (avan) {
          case avanzaA :
            nsa = nsa->siguiente;
	    fna = TRUE ;
	    break ;
          case avanzaB :
            nsb = nsb->siguiente;
	    fnb = TRUE ;
	    break ;
          case avanzanAyB :
            nsa = nsa->siguiente;
            nsb = nsb->siguiente;
	    fna = TRUE ;    fnb = TRUE ;
	    break ;
	  }
      }
    }
  *ncortes = cortes;
}
/**/
